@@ -6,7 +6,7 @@ use crate::hpack::{self, BytesStr};
 use http::header::{self, HeaderName, HeaderValue};
 use http::{uri, HeaderMap, Method, Request, StatusCode, Uri};
 
-use bytes::{BufMut, Bytes, BytesMut};
+use bytes::{Buf, BufMut, Bytes, BytesMut};
 
 use std::fmt;
 use std::io::Cursor;
@@ -166,7 +166,7 @@ impl Headers {
             pad = src[0] as usize;
 
             // Drop the padding
-            let _ = src.split_to(1);
+            src.advance(1);
         }
 
         // Read the stream dependency
@@ -181,7 +181,7 @@ impl Headers {
             }
 
             // Drop the next 5 bytes
-            let _ = src.split_to(5);
+            src.advance(5);
 
             Some(stream_dep)
         } else {
@@ -254,6 +254,10 @@ impl Headers {
         &mut self.header_block.pseudo
     }
 
+    pub(crate) fn pseudo(&self) -> &Pseudo {
+        &self.header_block.pseudo
+    }
+
     /// Whether it has status 1xx
     pub(crate) fn is_informational(&self) -> bool {
         self.header_block.pseudo.is_informational()
@@ -425,7 +429,7 @@ impl PushPromise {
             pad = src[0] as usize;
 
             // Drop the padding
-            let _ = src.split_to(1);
+            src.advance(1);
         }
 
         if src.len() < 5 {
@@ -434,7 +438,7 @@ impl PushPromise {
 
         let (promised_id, _) = StreamId::parse(&src[..4]);
         // Drop promised_id bytes
-        let _ = src.split_to(4);
+        src.advance(4);
 
         if pad > 0 {
             if pad > src.len() {
@@ -554,32 +558,36 @@ impl Pseudo {
     pub fn request(method: Method, uri: Uri, protocol: Option<Protocol>) -> Self {
         let parts = uri::Parts::from(uri);
 
-        let mut path = parts
-            .path_and_query
-            .map(|v| BytesStr::from(v.as_str()))
-            .unwrap_or(BytesStr::from_static(""));
+        let (scheme, path) = if method == Method::CONNECT && protocol.is_none() {
+            (None, None)
+        } else {
+            let path = parts
+                .path_and_query
+                .map(|v| BytesStr::from(v.as_str()))
+                .unwrap_or(BytesStr::from_static(""));
 
-        match method {
-            Method::OPTIONS | Method::CONNECT => {}
-            _ if path.is_empty() => {
-                path = BytesStr::from_static("/");
-            }
-            _ => {}
-        }
+            let path = if !path.is_empty() {
+                path
+            } else if method == Method::OPTIONS {
+                BytesStr::from_static("*")
+            } else {
+                BytesStr::from_static("/")
+            };
+
+            (parts.scheme, Some(path))
+        };
 
         let mut pseudo = Pseudo {
             method: Some(method),
             scheme: None,
             authority: None,
-            path: Some(path).filter(|p| !p.is_empty()),
+            path,
             protocol,
             status: None,
         };
 
         // If the URI includes a scheme component, add it to the pseudo headers
-        //
-        // TODO: Scheme must be set...
-        if let Some(scheme) = parts.scheme {
+        if let Some(scheme) = scheme {
             pseudo.set_scheme(scheme);
         }
 
@@ -653,7 +661,7 @@ impl EncodingHeaderBlock {
 
         // Now, encode the header payload
         let continuation = if self.hpack.len() > dst.remaining_mut() {
-            dst.put_slice(&self.hpack.split_to(dst.remaining_mut()));
+            dst.put((&mut self.hpack).take(dst.remaining_mut()));
 
             Some(Continuation {
                 stream_id: head.stream_id(),
@@ -1048,4 +1056,161 @@ mod test {
         let mut buf = BytesMut::new();
         huffman::decode(src, &mut buf).unwrap()
     }
+
+    #[test]
+    fn test_connect_request_pseudo_headers_omits_path_and_scheme() {
+        // CONNECT requests MUST NOT include :scheme & :path pseudo-header fields
+        // See: https://datatracker.ietf.org/doc/html/rfc9113#section-8.5
+
+        assert_eq!(
+            Pseudo::request(
+                Method::CONNECT,
+                Uri::from_static("https://example.com:8443"),
+                None
+            ),
+            Pseudo {
+                method: Method::CONNECT.into(),
+                authority: BytesStr::from_static("example.com:8443").into(),
+                ..Default::default()
+            }
+        );
+
+        assert_eq!(
+            Pseudo::request(
+                Method::CONNECT,
+                Uri::from_static("https://example.com/test"),
+                None
+            ),
+            Pseudo {
+                method: Method::CONNECT.into(),
+                authority: BytesStr::from_static("example.com").into(),
+                ..Default::default()
+            }
+        );
+
+        assert_eq!(
+            Pseudo::request(Method::CONNECT, Uri::from_static("example.com:8443"), None),
+            Pseudo {
+                method: Method::CONNECT.into(),
+                authority: BytesStr::from_static("example.com:8443").into(),
+                ..Default::default()
+            }
+        );
+    }
+
+    #[test]
+    fn test_extended_connect_request_pseudo_headers_includes_path_and_scheme() {
+        // On requests that contain the :protocol pseudo-header field, the
+        // :scheme and :path pseudo-header fields of the target URI (see
+        // Section 5) MUST also be included.
+        // See: https://datatracker.ietf.org/doc/html/rfc8441#section-4
+
+        assert_eq!(
+            Pseudo::request(
+                Method::CONNECT,
+                Uri::from_static("https://example.com:8443"),
+                Protocol::from_static("the-bread-protocol").into()
+            ),
+            Pseudo {
+                method: Method::CONNECT.into(),
+                authority: BytesStr::from_static("example.com:8443").into(),
+                scheme: BytesStr::from_static("https").into(),
+                path: BytesStr::from_static("/").into(),
+                protocol: Protocol::from_static("the-bread-protocol").into(),
+                ..Default::default()
+            }
+        );
+
+        assert_eq!(
+            Pseudo::request(
+                Method::CONNECT,
+                Uri::from_static("https://example.com:8443/test"),
+                Protocol::from_static("the-bread-protocol").into()
+            ),
+            Pseudo {
+                method: Method::CONNECT.into(),
+                authority: BytesStr::from_static("example.com:8443").into(),
+                scheme: BytesStr::from_static("https").into(),
+                path: BytesStr::from_static("/test").into(),
+                protocol: Protocol::from_static("the-bread-protocol").into(),
+                ..Default::default()
+            }
+        );
+
+        assert_eq!(
+            Pseudo::request(
+                Method::CONNECT,
+                Uri::from_static("http://example.com/a/b/c"),
+                Protocol::from_static("the-bread-protocol").into()
+            ),
+            Pseudo {
+                method: Method::CONNECT.into(),
+                authority: BytesStr::from_static("example.com").into(),
+                scheme: BytesStr::from_static("http").into(),
+                path: BytesStr::from_static("/a/b/c").into(),
+                protocol: Protocol::from_static("the-bread-protocol").into(),
+                ..Default::default()
+            }
+        );
+    }
+
+    #[test]
+    fn test_options_request_with_empty_path_has_asterisk_as_pseudo_path() {
+        // an OPTIONS request for an "http" or "https" URI that does not include a path component;
+        // these MUST include a ":path" pseudo-header field with a value of '*' (see Section 7.1 of [HTTP]).
+        // See: https://datatracker.ietf.org/doc/html/rfc9113#section-8.3.1
+        assert_eq!(
+            Pseudo::request(Method::OPTIONS, Uri::from_static("example.com:8080"), None,),
+            Pseudo {
+                method: Method::OPTIONS.into(),
+                authority: BytesStr::from_static("example.com:8080").into(),
+                path: BytesStr::from_static("*").into(),
+                ..Default::default()
+            }
+        );
+    }
+
+    #[test]
+    fn test_non_option_and_non_connect_requests_include_path_and_scheme() {
+        let methods = [
+            Method::GET,
+            Method::POST,
+            Method::PUT,
+            Method::DELETE,
+            Method::HEAD,
+            Method::PATCH,
+            Method::TRACE,
+        ];
+
+        for method in methods {
+            assert_eq!(
+                Pseudo::request(
+                    method.clone(),
+                    Uri::from_static("http://example.com:8080"),
+                    None,
+                ),
+                Pseudo {
+                    method: method.clone().into(),
+                    authority: BytesStr::from_static("example.com:8080").into(),
+                    scheme: BytesStr::from_static("http").into(),
+                    path: BytesStr::from_static("/").into(),
+                    ..Default::default()
+                }
+            );
+            assert_eq!(
+                Pseudo::request(
+                    method.clone(),
+                    Uri::from_static("https://example.com/a/b/c"),
+                    None,
+                ),
+                Pseudo {
+                    method: method.into(),
+                    authority: BytesStr::from_static("example.com").into(),
+                    scheme: BytesStr::from_static("https").into(),
+                    path: BytesStr::from_static("/a/b/c").into(),
+                    ..Default::default()
+                }
+            );
+        }
+    }
 }
