@@ -1,17 +1,17 @@
-use codec::{RecvError, UserError};
-use frame::{self, Reason};
 use super::{
-    store, Buffer, Codec, Config, Counts, Frame, Prioritize,
-    Prioritized, Store, Stream, StreamId, StreamIdOverflow, WindowSize,
+    store, Buffer, Codec, Config, Counts, Frame, Prioritize, Prioritized, Store, Stream, StreamId,
+    StreamIdOverflow, WindowSize,
 };
+use crate::codec::UserError;
+use crate::frame::{self, Reason};
+use crate::proto::{self, Error, Initiator};
 
 use bytes::Buf;
-use http;
-use futures::{Async, Poll};
-use futures::task::Task;
-use tokio_io::AsyncWrite;
+use tokio::io::AsyncWrite;
 
+use std::cmp::Ordering;
 use std::io;
+use std::task::{Context, Poll, Waker};
 
 /// Manages state transitions related to outbound frames.
 #[derive(Debug)]
@@ -19,11 +19,25 @@ pub(super) struct Send {
     /// Stream identifier to use for next initialized stream.
     next_stream_id: Result<StreamId, StreamIdOverflow>,
 
+    /// Any streams with a higher ID are ignored.
+    ///
+    /// This starts as MAX, but is lowered when a GOAWAY is received.
+    ///
+    /// > After sending a GOAWAY frame, the sender can discard frames for
+    /// > streams initiated by the receiver with identifiers higher than
+    /// > the identified last stream.
+    max_stream_id: StreamId,
+
     /// Initial window size of locally initiated streams
     init_window_sz: WindowSize,
 
     /// Prioritization layer
     prioritize: Prioritize,
+
+    is_push_enabled: bool,
+
+    /// If extended connect protocol is enabled.
+    is_extended_connect_protocol_enabled: bool,
 }
 
 /// A value to detect which public API has called `poll_reset`.
@@ -38,8 +52,11 @@ impl Send {
     pub fn new(config: &Config) -> Self {
         Send {
             init_window_sz: config.remote_init_window_sz,
+            max_stream_id: StreamId::MAX,
             next_stream_id: Ok(config.local_next_stream_id),
             prioritize: Prioritize::new(config),
+            is_push_enabled: true,
+            is_extended_connect_protocol_enabled: false,
         }
     }
 
@@ -54,56 +71,98 @@ impl Send {
         Ok(stream_id)
     }
 
+    pub fn reserve_local(&mut self) -> Result<StreamId, UserError> {
+        let stream_id = self.ensure_next_stream_id()?;
+        self.next_stream_id = stream_id.next_id();
+        Ok(stream_id)
+    }
+
+    fn check_headers(fields: &http::HeaderMap) -> Result<(), UserError> {
+        // 8.1.2.2. Connection-Specific Header Fields
+        if fields.contains_key(http::header::CONNECTION)
+            || fields.contains_key(http::header::TRANSFER_ENCODING)
+            || fields.contains_key(http::header::UPGRADE)
+            || fields.contains_key("keep-alive")
+            || fields.contains_key("proxy-connection")
+        {
+            tracing::debug!("illegal connection-specific headers found");
+            return Err(UserError::MalformedHeaders);
+        } else if let Some(te) = fields.get(http::header::TE) {
+            if te != "trailers" {
+                tracing::debug!("illegal connection-specific headers found");
+                return Err(UserError::MalformedHeaders);
+            }
+        }
+        Ok(())
+    }
+
+    pub fn send_push_promise<B>(
+        &mut self,
+        frame: frame::PushPromise,
+        buffer: &mut Buffer<Frame<B>>,
+        stream: &mut store::Ptr,
+        task: &mut Option<Waker>,
+    ) -> Result<(), UserError> {
+        if !self.is_push_enabled {
+            return Err(UserError::PeerDisabledServerPush);
+        }
+
+        tracing::trace!(
+            "send_push_promise; frame={:?}; init_window={:?}",
+            frame,
+            self.init_window_sz
+        );
+
+        Self::check_headers(frame.fields())?;
+
+        // Queue the frame for sending
+        self.prioritize
+            .queue_frame(frame.into(), buffer, stream, task);
+
+        Ok(())
+    }
+
     pub fn send_headers<B>(
         &mut self,
         frame: frame::Headers,
         buffer: &mut Buffer<Frame<B>>,
         stream: &mut store::Ptr,
         counts: &mut Counts,
-        task: &mut Option<Task>,
+        task: &mut Option<Waker>,
     ) -> Result<(), UserError> {
-        trace!(
+        tracing::trace!(
             "send_headers; frame={:?}; init_window={:?}",
             frame,
             self.init_window_sz
         );
 
-        // 8.1.2.2. Connection-Specific Header Fields
-        if frame.fields().contains_key(http::header::CONNECTION)
-            || frame.fields().contains_key(http::header::TRANSFER_ENCODING)
-            || frame.fields().contains_key(http::header::UPGRADE)
-            || frame.fields().contains_key("keep-alive")
-            || frame.fields().contains_key("proxy-connection")
-        {
-            debug!("illegal connection-specific headers found");
-            return Err(UserError::MalformedHeaders);
-        } else if let Some(te) = frame.fields().get(http::header::TE) {
-            if te != "trailers" {
-                debug!("illegal connection-specific headers found");
-                return Err(UserError::MalformedHeaders);
-
-            }
-        }
-
-        if frame.has_too_big_field() {
-            return Err(UserError::HeaderTooBig);
-        }
+        Self::check_headers(frame.fields())?;
 
         let end_stream = frame.is_end_stream();
 
         // Update the state
         stream.state.send_open(end_stream)?;
 
-        if counts.peer().is_local_init(frame.stream_id()) {
-            if counts.can_inc_num_send_streams() {
-                counts.inc_num_send_streams(stream);
-            } else {
-                self.prioritize.queue_open(stream);
-            }
+        let mut pending_open = false;
+        if counts.peer().is_local_init(frame.stream_id()) && !stream.is_pending_push {
+            self.prioritize.queue_open(stream);
+            pending_open = true;
         }
 
         // Queue the frame for sending
-        self.prioritize.queue_frame(frame.into(), buffer, stream, task);
+        //
+        // This call expects that, since new streams are in the open queue, new
+        // streams won't be pushed on pending_send.
+        self.prioritize
+            .queue_frame(frame.into(), buffer, stream, task);
+
+        // Need to notify the connection when pushing onto pending_open since
+        // queue_frame only notifies for pending_send.
+        if pending_open {
+            if let Some(task) = task.take() {
+                task.wake();
+            }
+        }
 
         Ok(())
     }
@@ -112,22 +171,25 @@ impl Send {
     pub fn send_reset<B>(
         &mut self,
         reason: Reason,
+        initiator: Initiator,
         buffer: &mut Buffer<Frame<B>>,
         stream: &mut store::Ptr,
         counts: &mut Counts,
-        task: &mut Option<Task>,
+        task: &mut Option<Waker>,
     ) {
         let is_reset = stream.state.is_reset();
         let is_closed = stream.state.is_closed();
         let is_empty = stream.pending_send.is_empty();
+        let stream_id = stream.id;
 
-        trace!(
-            "send_reset(..., reason={:?}, stream={:?}, ..., \
+        tracing::trace!(
+            "send_reset(..., reason={:?}, initiator={:?}, stream={:?}, ..., \
              is_reset={:?}; is_closed={:?}; pending_send.is_empty={:?}; \
              state={:?} \
-            ",
+             ",
             reason,
-            stream.id,
+            initiator,
+            stream_id,
             is_reset,
             is_closed,
             is_empty,
@@ -136,23 +198,23 @@ impl Send {
 
         if is_reset {
             // Don't double reset
-            trace!(
+            tracing::trace!(
                 " -> not sending RST_STREAM ({:?} is already reset)",
-                stream.id
+                stream_id
             );
             return;
         }
 
         // Transition the state to reset no matter what.
-        stream.state.set_reset(reason);
+        stream.set_reset(reason, initiator);
 
         // If closed AND the send queue is flushed, then the stream cannot be
         // reset explicitly, either. Implicit resets can still be queued.
         if is_closed && is_empty {
-            trace!(
+            tracing::trace!(
                 " -> not sending explicit RST_STREAM ({:?} was closed \
-                     and send queue was flushed)",
-                stream.id
+                 and send queue was flushed)",
+                stream_id
             );
             return;
         }
@@ -165,8 +227,9 @@ impl Send {
 
         let frame = frame::Reset::new(stream.id, reason);
 
-        trace!("send_reset -- queueing; frame={:?}", frame);
-        self.prioritize.queue_frame(frame.into(), buffer, stream, task);
+        tracing::trace!("send_reset -- queueing; frame={:?}", frame);
+        self.prioritize
+            .queue_frame(frame.into(), buffer, stream, task);
         self.prioritize.reclaim_all_capacity(stream, counts);
     }
 
@@ -175,7 +238,7 @@ impl Send {
         stream: &mut store::Ptr,
         reason: Reason,
         counts: &mut Counts,
-        task: &mut Option<Task>,
+        task: &mut Option<Waker>,
     ) {
         if stream.state.is_closed() {
             // Stream is already closed, nothing more to do
@@ -194,11 +257,13 @@ impl Send {
         buffer: &mut Buffer<Frame<B>>,
         stream: &mut store::Ptr,
         counts: &mut Counts,
-        task: &mut Option<Task>,
+        task: &mut Option<Waker>,
     ) -> Result<(), UserError>
-        where B: Buf,
+    where
+        B: Buf,
     {
-        self.prioritize.send_data(frame, buffer, stream, counts, task)
+        self.prioritize
+            .send_data(frame, buffer, stream, counts, task)
     }
 
     pub fn send_trailers<B>(
@@ -207,21 +272,18 @@ impl Send {
         buffer: &mut Buffer<Frame<B>>,
         stream: &mut store::Ptr,
         counts: &mut Counts,
-        task: &mut Option<Task>,
+        task: &mut Option<Waker>,
     ) -> Result<(), UserError> {
         // TODO: Should this logic be moved into state.rs?
         if !stream.state.is_send_streaming() {
             return Err(UserError::UnexpectedFrameType);
         }
 
-        if frame.has_too_big_field() {
-            return Err(UserError::HeaderTooBig);
-        }
-
         stream.state.send_close();
 
-        trace!("send_trailers -- queuing; frame={:?}", frame);
-        self.prioritize.queue_frame(frame.into(), buffer, stream, task);
+        tracing::trace!("send_trailers -- queuing; frame={:?}", frame);
+        self.prioritize
+            .queue_frame(frame.into(), buffer, stream, task);
 
         // Release any excess capacity
         self.prioritize.reserve_capacity(0, stream, counts);
@@ -231,15 +293,18 @@ impl Send {
 
     pub fn poll_complete<T, B>(
         &mut self,
+        cx: &mut Context,
         buffer: &mut Buffer<Frame<B>>,
         store: &mut Store,
         counts: &mut Counts,
         dst: &mut Codec<T, Prioritized<B>>,
-    ) -> Poll<(), io::Error>
-    where T: AsyncWrite,
-          B: Buf,
+    ) -> Poll<io::Result<()>>
+    where
+        T: AsyncWrite + Unpin,
+        B: Buf,
     {
-        self.prioritize.poll_complete(buffer, store, counts, dst)
+        self.prioritize
+            .poll_complete(cx, buffer, store, counts, dst)
     }
 
     /// Request capacity to send data
@@ -247,52 +312,47 @@ impl Send {
         &mut self,
         capacity: WindowSize,
         stream: &mut store::Ptr,
-        counts: &mut Counts)
-    {
+        counts: &mut Counts,
+    ) {
         self.prioritize.reserve_capacity(capacity, stream, counts)
     }
 
     pub fn poll_capacity(
         &mut self,
+        cx: &Context,
         stream: &mut store::Ptr,
-    ) -> Poll<Option<WindowSize>, UserError> {
+    ) -> Poll<Option<Result<WindowSize, UserError>>> {
         if !stream.state.is_send_streaming() {
-            return Ok(Async::Ready(None));
+            return Poll::Ready(None);
         }
 
         if !stream.send_capacity_inc {
-            stream.wait_send();
-            return Ok(Async::NotReady);
+            stream.wait_send(cx);
+            return Poll::Pending;
         }
 
         stream.send_capacity_inc = false;
 
-        Ok(Async::Ready(Some(self.capacity(stream))))
+        Poll::Ready(Some(Ok(self.capacity(stream))))
     }
 
     /// Current available stream send capacity
     pub fn capacity(&self, stream: &mut store::Ptr) -> WindowSize {
-        let available = stream.send_flow.available().as_size();
-        let buffered = stream.buffered_send_data;
-
-        if available <= buffered {
-            0
-        } else {
-            available - buffered
-        }
+        stream.capacity(self.prioritize.max_buffer_size())
     }
 
     pub fn poll_reset(
         &self,
+        cx: &Context,
         stream: &mut Stream,
         mode: PollReset,
-    ) -> Poll<Reason, ::Error> {
+    ) -> Poll<Result<Reason, crate::Error>> {
         match stream.state.ensure_reason(mode)? {
-            Some(reason) => Ok(reason.into()),
+            Some(reason) => Poll::Ready(Ok(reason)),
             None => {
-                stream.wait_send();
-                Ok(Async::NotReady)
-            },
+                stream.wait_send(cx);
+                Poll::Pending
+            }
         }
     }
 
@@ -312,14 +372,19 @@ impl Send {
         buffer: &mut Buffer<Frame<B>>,
         stream: &mut store::Ptr,
         counts: &mut Counts,
-        task: &mut Option<Task>,
+        task: &mut Option<Waker>,
     ) -> Result<(), Reason> {
         if let Err(e) = self.prioritize.recv_stream_window_update(sz, stream) {
-            debug!("recv_stream_window_update !!; err={:?}", e);
+            tracing::debug!("recv_stream_window_update !!; err={:?}", e);
 
             self.send_reset(
-                Reason::FLOW_CONTROL_ERROR.into(),
-                buffer, stream, counts, task);
+                Reason::FLOW_CONTROL_ERROR,
+                Initiator::Library,
+                buffer,
+                stream,
+                counts,
+                task,
+            );
 
             return Err(e);
         }
@@ -327,7 +392,27 @@ impl Send {
         Ok(())
     }
 
-    pub fn recv_err<B>(
+    pub(super) fn recv_go_away(&mut self, last_stream_id: StreamId) -> Result<(), Error> {
+        if last_stream_id > self.max_stream_id {
+            // The remote endpoint sent a `GOAWAY` frame indicating a stream
+            // that we never sent, or that we have already terminated on account
+            // of previous `GOAWAY` frame. In either case, that is illegal.
+            // (When sending multiple `GOAWAY`s, "Endpoints MUST NOT increase
+            // the value they send in the last stream identifier, since the
+            // peers might already have retried unprocessed requests on another
+            // connection.")
+            proto_err!(conn:
+                "recv_go_away: last_stream_id ({:?}) > max_stream_id ({:?})",
+                last_stream_id, self.max_stream_id,
+            );
+            return Err(Error::library_go_away(Reason::PROTOCOL_ERROR));
+        }
+
+        self.max_stream_id = last_stream_id;
+        Ok(())
+    }
+
+    pub fn handle_error<B>(
         &mut self,
         buffer: &mut Buffer<Frame<B>>,
         stream: &mut store::Ptr,
@@ -344,8 +429,12 @@ impl Send {
         buffer: &mut Buffer<Frame<B>>,
         store: &mut Store,
         counts: &mut Counts,
-        task: &mut Option<Task>,
-    ) -> Result<(), RecvError> {
+        task: &mut Option<Waker>,
+    ) -> Result<(), Error> {
+        if let Some(val) = settings.is_extended_connect_protocol_enabled() {
+            self.is_extended_connect_protocol_enabled = val;
+        }
+
         // Applies an update to the remote endpoint's initial window size.
         //
         // Per RFC 7540 §6.9.2:
@@ -367,62 +456,84 @@ impl Send {
             let old_val = self.init_window_sz;
             self.init_window_sz = val;
 
-            if val < old_val {
-                // We must decrease the (remote) window on every open stream.
-                let dec = old_val - val;
-                trace!("decrementing all windows; dec={}", dec);
+            match val.cmp(&old_val) {
+                Ordering::Less => {
+                    // We must decrease the (remote) window on every open stream.
+                    let dec = old_val - val;
+                    tracing::trace!("decrementing all windows; dec={}", dec);
 
-                let mut total_reclaimed = 0;
-                store.for_each(|mut stream| {
-                    let stream = &mut *stream;
+                    let mut total_reclaimed = 0;
+                    store.try_for_each(|mut stream| {
+                        let stream = &mut *stream;
 
-                    stream.send_flow.dec_window(dec);
+                        tracing::trace!(
+                            "decrementing stream window; id={:?}; decr={}; flow={:?}",
+                            stream.id,
+                            dec,
+                            stream.send_flow
+                        );
 
-                    // It's possible that decreasing the window causes
-                    // `window_size` (the stream-specific window) to fall below
-                    // `available` (the portion of the connection-level window
-                    // that we have allocated to the stream).
-                    // In this case, we should take that excess allocation away
-                    // and reassign it to other streams.
-                    let window_size = stream.send_flow.window_size();
-                    let available = stream.send_flow.available().as_size();
-                    let reclaimed = if available > window_size {
-                        // Drop down to `window_size`.
-                        let reclaim = available - window_size;
-                        stream.send_flow.claim_capacity(reclaim);
-                        total_reclaimed += reclaim;
-                        reclaim
-                    } else {
-                        0
-                    };
+                        // TODO: this decrement can underflow based on received frames!
+                        stream
+                            .send_flow
+                            .dec_send_window(dec)
+                            .map_err(proto::Error::library_go_away)?;
 
-                    trace!(
-                        "decremented stream window; id={:?}; decr={}; reclaimed={}; flow={:?}",
-                        stream.id,
-                        dec,
-                        reclaimed,
-                        stream.send_flow
-                    );
+                        // It's possible that decreasing the window causes
+                        // `window_size` (the stream-specific window) to fall below
+                        // `available` (the portion of the connection-level window
+                        // that we have allocated to the stream).
+                        // In this case, we should take that excess allocation away
+                        // and reassign it to other streams.
+                        let window_size = stream.send_flow.window_size();
+                        let available = stream.send_flow.available().as_size();
+                        let reclaimed = if available > window_size {
+                            // Drop down to `window_size`.
+                            let reclaim = available - window_size;
+                            stream
+                                .send_flow
+                                .claim_capacity(reclaim)
+                                .map_err(proto::Error::library_go_away)?;
+                            total_reclaimed += reclaim;
+                            reclaim
+                        } else {
+                            0
+                        };
 
-                    // TODO: Should this notify the producer when the capacity
-                    // of a stream is reduced? Maybe it should if the capacity
-                    // is reduced to zero, allowing the producer to stop work.
+                        tracing::trace!(
+                            "decremented stream window; id={:?}; decr={}; reclaimed={}; flow={:?}",
+                            stream.id,
+                            dec,
+                            reclaimed,
+                            stream.send_flow
+                        );
 
-                    Ok::<_, RecvError>(())
-                })?;
+                        // TODO: Should this notify the producer when the capacity
+                        // of a stream is reduced? Maybe it should if the capacity
+                        // is reduced to zero, allowing the producer to stop work.
 
-                self.prioritize
-                    .assign_connection_capacity(total_reclaimed, store, counts);
-            } else if val > old_val {
-                let inc = val - old_val;
+                        Ok::<_, proto::Error>(())
+                    })?;
 
-                store.for_each(|mut stream| {
-                    self.recv_stream_window_update(inc, buffer, &mut stream, counts, task)
-                        .map_err(RecvError::Connection)
-                })?;
+                    self.prioritize
+                        .assign_connection_capacity(total_reclaimed, store, counts);
+                }
+                Ordering::Greater => {
+                    let inc = val - old_val;
+
+                    store.try_for_each(|mut stream| {
+                        self.recv_stream_window_update(inc, buffer, &mut stream, counts, task)
+                            .map_err(Error::library_go_away)
+                    })?;
+                }
+                Ordering::Equal => (),
             }
         }
 
+        if let Some(val) = settings.is_push_enabled() {
+            self.is_push_enabled = val
+        }
+
         Ok(())
     }
 
@@ -444,19 +555,31 @@ impl Send {
     }
 
     pub fn ensure_next_stream_id(&self) -> Result<StreamId, UserError> {
-        self.next_stream_id.map_err(|_| UserError::OverflowedStreamId)
+        self.next_stream_id
+            .map_err(|_| UserError::OverflowedStreamId)
     }
 
     pub fn may_have_created_stream(&self, id: StreamId) -> bool {
         if let Ok(next_id) = self.next_stream_id {
             // Peer::is_local_init should have been called beforehand
-            debug_assert_eq!(
-                id.is_server_initiated(),
-                next_id.is_server_initiated(),
-            );
+            debug_assert_eq!(id.is_server_initiated(), next_id.is_server_initiated(),);
             id < next_id
         } else {
             true
         }
     }
+
+    pub(super) fn maybe_reset_next_stream_id(&mut self, id: StreamId) {
+        if let Ok(next_id) = self.next_stream_id {
+            // Peer::is_local_init should have been called beforehand
+            debug_assert_eq!(id.is_server_initiated(), next_id.is_server_initiated());
+            if id >= next_id {
+                self.next_stream_id = id.next_id();
+            }
+        }
+    }
+
+    pub(crate) fn is_extended_connect_protocol_enabled(&self) -> bool {
+        self.is_extended_connect_protocol_enabled
+    }
 }
